use crate::data::id::ID;
use std::{collections::HashMap, fmt, ops::Deref};

#[derive(Clone)]
pub enum Field {
    ID(ID),
}

impl From<ID> for Field {
    fn from(value: ID) -> Self {
        Field::ID(value)
    }
}

impl fmt::Display for Field {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Field::ID(id) => write!(f, "{}", id),
        }
    }
}

pub struct Record {
    data: HashMap<String, Field>,
}

impl Record {
    pub fn new() -> Self {
        Self {
            data: HashMap::new(),
        }
    }

    pub fn add<S, D>(&mut self, name: S, data: D) -> Option<Field>
    where
        String: From<S>,
        Field: From<D>,
    {
        self.data.insert(name.into(), data.into())
    }
}

impl Deref for Record {
    type Target = HashMap<String, Field>;

    fn deref(&self) -> &Self::Target {
        &self.data
    }
}

#[cfg(test)]
mod fields {
    use super::*;

    #[test]
    fn from_id() {
        let id = ID::random();
        let field = Field::from(id.clone());
        assert_eq!(field.to_string(), id.to_string());
    }
}

#[cfg(test)]
mod records {
    use super::*;

    #[test]
    fn new_record() {
        let rec = Record::new();
        assert_eq!(rec.len(), 0);
    }

    #[test]
    fn add_field() {
        let mut rec = Record::new();
        let name = "field_name";
        let data = ID::random();
        rec.add(name.to_string(), data);
        match rec.get(name) {
            Some(data) => match data {
                Field::ID(result) => assert_eq!(result.to_string(), data.to_string()),
            },
            None => unreachable!("Should get data back"),
        }
    }

    #[test]
    fn add_alternate_field() {
        let mut rec = Record::new();
        let name = "alternate";
        let data = ID::random();
        rec.add(name, data.clone());
        match rec.get(name) {
            Some(data) => match data {
                Field::ID(result) => assert_eq!(result.to_string(), data.to_string()),
            },
            None => unreachable!("Should get data back"),
        }
    }

    #[test]
    fn record_iter() {
        let mut data: HashMap<String, ID> = HashMap::new();
        data.insert("a".to_string(), ID::random());
        data.insert("b".to_string(), ID::random());
        data.insert("c".to_string(), ID::random());
        let mut rec = Record::new();
        for (key, value) in data.iter() {
            rec.add(key, value.clone());
        }
        assert_eq!(rec.len(), data.len());
        for (key, value) in rec.iter() {
            assert_eq!(value.to_string(), data.get(key).unwrap().to_string());
        }
    }
}
